from thriftpy2.thrift import TType


class ThriftError(Exception):
    """ Base Exception defined by `aiothrift` """


class ConnectionClosedError(ThriftError):
    """Raised if connection to server was closed."""


class PoolClosedError(ThriftError):
    """Raised when operating on a closed thrift connection pool"""


class ThriftAppError(ThriftError):
    """Application level thrift exceptions."""

    thrift_spec = {
        1: (TType.STRING, "message", False),
        2: (TType.I32, "type", False),
    }

    UNKNOWN = 0
    UNKNOWN_METHOD = 1
    INVALID_MESSAGE_TYPE = 2
    WRONG_METHOD_NAME = 3
    BAD_SEQUENCE_ID = 4
    MISSING_RESULT = 5
    INTERNAL_ERROR = 6
    PROTOCOL_ERROR = 7

    def __init__(self, type=UNKNOWN, message=None):
        super().__init__()
        self.type = type
        self.message = message

    def __str__(self):
        if self.message:
            return self.message

        if self.type == self.UNKNOWN_METHOD:
            return "Unknown method"
        elif self.type == self.INVALID_MESSAGE_TYPE:
            return "Invalid message type"
        elif self.type == self.WRONG_METHOD_NAME:
            return "Wrong method name"
        elif self.type == self.BAD_SEQUENCE_ID:
            return "Bad sequence ID"
        elif self.type == self.MISSING_RESULT:
            return "Missing result"
        else:
            return "Default (unknown) TApplicationException"



from thriftpy2 import load
from .connection import create_connection, ThriftConnection
from .pool import create_pool, ThriftPool
from .errors import ConnectionClosedError, ThriftError, PoolClosedError, ThriftAppError
from .processor import TProcessor
from .server import Server, create_server
from .protocol import TBinaryProtocol, TProtocol

__version__ = "0.2.0"

__all__ = [
    "create_connection",
    "ThriftAppError",
    "ThriftConnection",
    "create_pool",
    "ThriftError",
    "TBinaryProtocol",
    "TProcessor",
    "TProtocol",
    "ThriftPool",
    "ConnectionClosedError",
    "PoolClosedError",
    "Server",
    "create_server",
    "load",
]



import asyncio
import functools

import async_timeout
from thriftpy2.thrift import TMessageType

from .protocol import TBinaryProtocol
from .util import args2kwargs
from .errors import ConnectionClosedError, ThriftAppError
from .log import logger


async def create_connection(
    service,
    address=("127.0.0.1", 6000),
    *,
    protocol_cls=TBinaryProtocol,
    timeout=None,
    **kw,
):
    """Create a thrift connection.
    This function is a :ref:`coroutine <coroutine>`.

    Open a connection to the thrift server by address argument.

    :param service: a thrift service object
    :param address: a (host, port) tuple
    :param protocol_cls: protocol type, default is :class:`TBinaryProtocol`
    :param timeout: if specified, would raise `asyncio.TimeoutError` if one rpc call
    is longer than `timeout`
    :param kw: params relaied to asyncio.open_connection()
    :return: newly created :class:`ThriftConnection` instance.
    """
    host, port = address
    reader, writer = await asyncio.open_connection(host, port, **kw)
    iprotocol = protocol_cls(reader)
    oprotocol = protocol_cls(writer)

    return ThriftConnection(
        service, iprot=iprotocol, oprot=oprotocol, address=address, timeout=timeout
    )


class ThriftConnection:
    """
    Thrift Connection.
    """

    def __init__(self, service, *, iprot, oprot, address, timeout=None):
        self.service = service
        self._reader = iprot.trans
        self._writer = oprot.trans
        self.timeout = timeout
        self.address = address
        self.closed = False
        self._oprot = oprot
        self._iprot = iprot
        self._seqid = 0
        self._init_rpc_apis()

    def _init_rpc_apis(self):
        """
        find out all apis defined in thrift service, and create corresponding
        method on the connection object, ignore it if some api name is conflicted with
        an existed attribute of the connection object, which you should call by using
        the :meth:`execute` method.
        """
        for api in self.service.thrift_services:
            if not hasattr(self, api):
                setattr(self, api, functools.partial(self.execute, api))
            else:
                logger.warning(
                    "api name {0} is conflicted with connection attribute "
                    '{0}, while you can still call this api by `send_call("{0}")`'.format(
                        api
                    )
                )

    def __repr__(self):
        return "<ThriftConnection {} to>".format(self.address)

    async def execute(self, api, *args, **kwargs):
        """
        Execute a rpc call by api name. This is function is a :ref:`coroutine <coroutine>`.

        :param api: api name defined in thrift file
        :param args: positional arguments passed to api function
        :param kwargs:  keyword arguments passed to api function
        :return: result of this rpc call
        :raises: :class:`~asyncio.TimeoutError` if this task has exceeded the `timeout`
        :raises: :class:`ThriftAppError` if thrift response is an exception defined in thrift.
        :raises: :class:`ConnectionClosedError`: if server has closed this connection.
        """
        if self.closed:
            raise ConnectionClosedError("Connection closed")

        try:
            with async_timeout.timeout(self.timeout):
                kw = args2kwargs(
                    getattr(self.service, api + "_args").thrift_spec, *args
                )
                kwargs.update(kw)
                result_cls = getattr(self.service, api + "_result")

                self._seqid += 1
                self._oprot.write_message_begin(api, TMessageType.CALL, self._seqid)
                args = getattr(self.service, api + "_args")()
                for k, v in kwargs.items():
                    setattr(args, k, v)
                args.write(self._oprot)
                self._oprot.write_message_end()
                await self._oprot.trans.drain()
                if not getattr(result_cls, "oneway"):
                    result = await self._recv(api)
                    return result
        except asyncio.TimeoutError:
            self.close()
            raise
        except ConnectionError as e:
            self.close()
            logger.debug("connection error {}".format(str(e)))
            raise ConnectionClosedError("the server has closed this connection") from e
        except asyncio.IncompleteReadError as e:
            self.close()
            raise ConnectionClosedError("Server connection has closed") from e

    async def _recv(self, api):
        """
        A :ref:`coroutine <coroutine>` which receive response from the thrift server
        """
        fname, mtype, rseqid = await self._iprot.read_message_begin()
        if rseqid != self._seqid:
            # transport should be closed if bad seq happened
            self.close()
            raise ThriftAppError(
                ThriftAppError.BAD_SEQUENCE_ID,
                fname + " failed: out of sequence response",
            )

        if mtype == TMessageType.EXCEPTION:
            x = ThriftAppError()
            await self._iprot.read_struct(x)
            await self._iprot.read_message_end()
            raise x
        result = getattr(self.service, api + "_result")()
        await self._iprot.read_struct(result)
        await self._iprot.read_message_end()

        if hasattr(result, "success") and result.success is not None:
            return result.success

        # void api without throws
        if len(result.thrift_spec) == 0:
            return

        # check throws
        for k, v in result.__dict__.items():
            if k != "success" and v:
                raise v
        if hasattr(result, "success"):
            raise ThriftAppError(ThriftAppError.MISSING_RESULT)

    def close(self):
        self._writer.close()
        self.closed = True



"""
The connection pool implementation is heavily borrowed from `aioredis`
"""
import asyncio
import contextvars as cv
import collections

from .connection import create_connection
from .log import logger
from .errors import PoolClosedError


acquired_connection = cv.ContextVar("acquired_connection")


async def create_pool(
    service, address=("127.0.0.1", 6000), *, minsize=1, maxsize=10, timeout=None
):
    """
    Create a thrift connection pool. This function is a :ref:`coroutine <coroutine>`.

    :param service: service object defined by thrift file
    :param address: (host, port) tuple, default is ('127.0.0.1', 6000)
    :param minsize: minimal thrift connection, default is 1
    :param maxsize: maximal thrift connection, default is 10
    :param timeout: default timeout for each connection, default is None
    :return: :class:`ThriftPool` instance
    """

    pool = ThriftPool(
        service, address, minsize=minsize, maxsize=maxsize, timeout=timeout
    )
    try:
        await pool.fill_free(override_min=False)
    except Exception:
        pool.close()
        raise

    return pool


class ThriftPool:
    """Thrift connection pool.
    """

    def __init__(self, service, address, *, minsize, maxsize, timeout=None):
        assert isinstance(minsize, int) and minsize >= 0, (
            "minsize must be int >= 0",
            minsize,
            type(minsize),
        )
        assert maxsize is not None, "Arbitrary pool size is disallowed."
        assert isinstance(maxsize, int) and maxsize > 0, (
            "maxsize must be int > 0",
            maxsize,
            type(maxsize),
        )
        assert minsize <= maxsize, ("Invalid pool min/max sizes", minsize, maxsize)
        self._address = address
        self.minsize = minsize
        self.maxsize = maxsize
        self._pool = collections.deque(maxlen=maxsize)
        self._used = set()
        self._acquiring = 0
        self._cond = asyncio.Condition()
        self._service = service
        self._timeout = timeout
        self.closed = False
        self._release_tasks = set()

    @property
    def size(self):
        """Current connection total num, acquiring connection num is counted"""
        return self.freesize + len(self._used) + self._acquiring

    @property
    def freesize(self):
        """Current number of free connections."""
        return len(self._pool)

    async def clear(self):
        """Clear pool connections.

        Close and remove all free connections.
        this pattern is interesting
        """
        while self._pool:
            conn = self._pool.popleft()
            conn.close()

    def close(self):
        """Close all free and in-progress connections and mark pool as closed.
        """
        self.closed = True
        conn_num = 0
        while self._pool:
            conn = self._pool.popleft()
            conn.close()
            conn_num += 1
        for conn in self._used:
            conn.close()
            conn_num += 1
        logger.debug("Closed %d connections", conn_num)

    async def wait_closed(self):
        for task in self._release_tasks:
            await asyncio.shield(task)

    async def acquire(self):
        """Acquires a connection from free pool.

        Creates new connection if needed.
        """
        if self.closed:
            raise PoolClosedError("Pool is closed")
        async with self._cond:
            if self.closed:
                raise PoolClosedError("Pool is closed")
            while True:
                await self.fill_free(override_min=True)
                # new connection has been added to the pool
                if self.freesize:
                    conn = self._pool.popleft()
                    assert not conn.closed, conn
                    assert conn not in self._used, (conn, self._used)
                    # each acquire would move a conn from `self._pool` to `self._used`
                    self._used.add(conn)
                    return conn
                else:
                    # wait when no available connection
                    await self._cond.wait()

    def release(self, conn):
        """Returns used connection back into pool.

        When queue of free connections is full the connection will be dropped.
        """
        assert conn in self._used, "Invalid connection, maybe from other pool"
        self._used.remove(conn)
        if not conn.closed:
            assert self.freesize < self.maxsize, "max connection size should not exceed"
            self._pool.append(conn)

        loop = asyncio.get_running_loop()
        if not loop.is_closed():
            tasks = set()
            for task in self._release_tasks:
                if not task.done():
                    tasks.add(task)
            self._release_tasks = tasks
            future = asyncio.create_task(self._notify_conn_returned())
            self._release_tasks.add(future)

    def _drop_closed(self):
        for i in range(self.freesize):
            conn = self._pool[0]
            if conn.closed:
                self._pool.popleft()
            else:
                self._pool.rotate(1)

    async def fill_free(self, *, override_min):
        """
        make sure at least `self.minsize` amount of connections in the pool
        if `override_min` is True, fill to the `self.maxsize`.
        """
        # drop closed connections first, in case that the user closed the connection manually
        self._drop_closed()

        while self.size < self.minsize:
            self._acquiring += 1
            try:
                conn = await self._create_new_connection()
                self._pool.append(conn)
            finally:
                self._acquiring -= 1
                self._drop_closed()
        if self.freesize:
            return
        if override_min:
            # when self.size >= minsize and no available connection
            while not self._pool and self.size < self.maxsize:
                self._acquiring += 1
                try:
                    conn = await self._create_new_connection()
                    self._pool.append(conn)
                finally:
                    self._acquiring -= 1

    def _create_new_connection(self):
        return create_connection(self._service, self._address, timeout=self._timeout)

    async def _notify_conn_returned(self):
        async with self._cond:
            self._cond.notify()

    async def __aenter__(self):
        if self.closed:
            raise PoolClosedError("cannot acquire a connection from a closed pool")
        if acquired_connection.get(None) is not None:
            raise RuntimeError(
                "cannot acquire a connection if you already have one inside the same task"
            )
        _conn = await self.acquire()
        acquired_connection.set(_conn)
        return _conn

    async def __aexit__(self, *exc_info):
        self.release(acquired_connection.get())



import asyncio

import async_timeout

from .log import logger
from .processor import TProcessor
from .protocol import TBinaryProtocol


class Server:
    def __init__(self, processor, protocol_cls=TBinaryProtocol, timeout=None):
        self.processor = processor
        self.protocol_cls = protocol_cls
        self.timeout = timeout

    async def __call__(self, reader, writer):
        iproto = self.protocol_cls(reader)
        oproto = self.protocol_cls(writer)
        while not reader.at_eof():
            try:
                with async_timeout.timeout(self.timeout):
                    await self.processor.process(iproto, oproto)
            except ConnectionError:
                logger.debug("client has closed the connection")
                writer.close()
            except asyncio.TimeoutError:
                logger.debug("timeout when processing the client request")
                writer.close()
            except asyncio.IncompleteReadError:
                logger.debug("client has closed the connection")
                writer.close()
            except Exception:
                # app exception
                logger.exception("unhandled app exception")
                writer.close()
        writer.close()


async def create_server(
    service,
    handler,
    address=("127.0.0.1", 6000),
    protocol_cls=TBinaryProtocol,
    timeout=None,
    **kw,
):
    """ create a thrift server.
    This function is a :ref:`coroutine <coroutine>`.

    :param service: thrift Service
    :param handler: a dispatcher object which is a namespace for all thrift api functions.
    :param address:  (host, port) tuple, default is ('127.0.0.1', 6000)
    :param protocol_cls: thrift protocol class, default is :class:`TBinaryProtocol`
    :param timeout: server side timeout, default is None
    :param kw: params relaied to asyncio.start_server
    :return: a :class:`Server` object which can be used to stop the service
    """
    host, port = address
    processor = TProcessor(service, handler)
    server = await asyncio.start_server(
        Server(processor, protocol_cls, timeout=timeout), host, port, **kw
    )
    return server



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiothrift']

package_data = \
{'': ['*']}

install_requires = \
['async-timeout>=3.0.1,<4.0.0', 'thriftpy2>=0.4.9,<0.5.0']

setup_kwargs = {
    'name': 'aiothrift',
    'version': '0.2.0',
    'description': 'Async Thrift server and client',
    'long_description': 'aiothrift\n=========\n\nAsyncio implementation for thrift protocol, which is heavily based on thriftpy_.\n\n.. image:: https://travis-ci.org/ryanwang520/aiothrift.svg?branch=master\n   :target: https://travis-ci.org/ryanwang520/aiothrift\n\n\nDocumentation: https://aiothrift.readthedocs.org/\n\nInstallation\n------------\n\n::\n\n    $ pip install aiothrift\n\n\nUsage example\n-------------\n\nThrift file\n^^^^^^^^^^^\n\n::\n\n    service PingPong {\n        string ping(),\n        i64 add(1:i32 a, 2:i64 b),\n    }\n\n\nServer\n^^^^^^\n\n.. code:: python\n\n    import asyncio\n    import aiothrift\n\n    pingpong_thrift = aiothrift.load(\'pingpong.thrift\', module_name=\'pingpong_thrift\')\n\n    class Dispatcher:\n        def ping(self):\n            return "pong"\n\n        async def add(self, a, b):\n            await asyncio.sleep(1)\n            return a + b\n\n    async def main():\n      server = await aiothrift.create_server(pingpong_thrift.PingPong, Dispatcher()))\n      async with server:\n          await server.serve_forever()\n    asyncio.run(main())\n\nClient\n^^^^^^\n\n.. code:: python\n\n    import asyncio\n    import aiothrift\n\n    pingpong_thrift = aiothrift.load(\'pingpong.thrift\', module_name=\'pingpong_thrift\')\n\n    async def go():\n        conn = await aiothrift.create_connection(pingpong_thrift.PingPong)\n        print(await conn.ping())\n        print(await conn.add(5, 6))\n        conn.close()\n\n    asyncio.run(go)\n\nOr use ConnectionPool\n^^^^^^^^^^^^^^^^^^^^^\n\n.. code:: python\n\n    import asyncio\n    import aiothrift\n\n    pingpong_thrift = aiothrift.load(\'pingpong.thrift\', module_name=\'pingpong_thrift\')\n\n    async def go():\n        pool = await aiothrift.create_pool(pingpong_thrift.PingPong)\n        async with pool as c:\n          print(await c.ping())\n          print(await c.add(5, 6))\n        pool.close()\n        await pool.wait_closed()\n\n    asyncio.run(go())\n\n\nIt\'s just that simple to begin with ``aiothrift``, and you are not forced to use ``aiothrift`` on both server and client side.\nSo if you already have a normal thrift server setup, feel free to create an async thrift client to communicate with that server.\n\nRequirements\n------------\n\n- Python >= 3.4.2\n- async-timeout_\n- thriftpy_\n\n.. _async-timeout: https://pypi.python.org/pypi/async_timeout\n.. _thriftpy: https://thriftpy.readthedocs.org/en/latest/\n\n\nLICENSE\n-------\n\n``aiothrift`` is offered under the MIT license.\n',
    'author': 'Ryan Wang',
    'author_email': 'hwwangwang@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://github.com/moonshadow/aiothrift/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
